//	Initialize the copyWidget object
prism.run([

    "$filter",
    "plugin-RegressionLine.services.bindOnce",
    "plugin-RegressionLine.services.regressionFunc",

    function($filter, $bindOnce,$regressionFunc) {

	//////////////////////////
	//	Plugin properties	//
	//////////////////////////

	//	Regression options
	mod.none = 'none';

	//var regressionOptions = [none,'linear','exponential','polynomial','logarithmic','loess'];
	var regressionOptions = [mod.none, 'linear','exponential','logarithmic'];
	mod.regressionProperty = 'regressionType';	

	//	Prediciton methods for how many
	var calendarOption = 'calendar end';
	//var predictionMethods = [none, calendarOption,'rolling 3', 'rolling 6', 'rolling 12'];
	mod.predictionMethods = [
		{
			label: mod.none,
		},
		{
			label: calendarOption
		},
		{
			label: 'Next 3',
			numPeriods: 3
		},
		{
			label: 'Next 6',
			numPeriods: 6
		},
		{
			label: 'Next 12',
			numPeriods: 12
		},
	];

	//	Default settings
	mod.defaultSettings = {
		enabled : false,
		predictionMethod: calendarOption,
        periodsAdded:false,
	};

	//	Define future periods
	var yearsToPredict = 5,
		weeksToPredict = 5;

	//	Get the date formatter
	prism.dateFormatter = $filter("levelDate");		

	//////////////////////////////////
	//	Initialization Functions	//
	//////////////////////////////////

	// Registering dashboard/ widget creation in order to perform drilling
	prism.on("dashboardloaded", function (e, args) {
        var widgets = args.dashboard.widgets.$$widgets;
        for (var i = 0; i < widgets.length; i++) {
            //	Hooking to ready/destroyed events
            $bindOnce(widgets[i], "destroyed" , onWidgetDestroyed);
            $bindOnce(widgets[i], "processresult", $regressionFunc.startRegression);
        }
    });

	// unregistering widget events
	function onWidgetDestroyed(widget, args) {
		widget.off("destroyed", onWidgetDestroyed);
	}

	//	Create Options for the quadrant analysis
	prism.on("beforemenu", function (e, args) {

		//	Define the menu label
		var headerMenuCaption = "Regression Line";

		//	Can we show the regression options?			
		var addMeasureMenuItem = $regressionFunc.canEnableRegression(e,args,headerMenuCaption,'measure');
		var addSettingsMenuItem = $regressionFunc.canEnableRegression(e,args,headerMenuCaption,'settings');

		//	Add the option to enable the regression line
		if (addMeasureMenuItem) {			
			
			//	Get the widget
			var widget = args.settings.item.$$panel.$$widget;	

			//	Look for a saved regression property
			var savedRegressionType = args.settings.item[mod.regressionProperty] ? args.settings.item[mod.regressionProperty] : mod.none;

			//	Function to create the measure sub-menu items				
			var regressionTypeItems = function(panelItem) {
				//	Create an array to hold the list
				var items = [];

				//	Get the setting for this
				var selectedType = savedRegressionType;
				
				//	Function the runs when an item is picked
				var itemPicked = function(){						
					//	Update the axis setting
					this.panelItem[mod.regressionProperty] = this.caption;
					//	Redraw the widget
					widget.redraw();						
				};

				//	Loop through each regression option and add sub-menu item
				$.each(regressionOptions,function(){
					
					//	Get the option as a string
					var option = this.toString();

					//	Create the menu option
					var item = {
						caption: option,
						checked: (option === selectedType),
						size: 'xl',
						type: 'check',
						panelItem: panelItem,
						execute: itemPicked
					};

					//	Add to the list
					items.push(item);
				});		

				//	Return the list of items				
				return items;
			};

			//	Create menu for picking the regression measure
			var regressionTypes = {
				caption: "Regression",
				items: regressionTypeItems(args.settings.item)
			};

			//	Add options to the menu			
			args.settings.items.push(regressionTypes);	
		}

		//	Add the options in the settings menu
		if (addSettingsMenuItem) {

			//	Get the widget
			var widget = e.currentScope.widget;

			//	Get saved settings
			var settings = widget.options.regression;
			if (typeof settings === 'undefined'){
				//	Create defaults, if not defined
				settings = mod.defaultSettings;
			}

			//	Function to create the axis menu items				
			var predictionOptionItems = function(settings) {
				//	Create an array to hold the list
				var items = [];

				//	Get the setting for this				
				var selectedMethod = settings.predictionMethod;
				
				//	Function the runs when an item is picked
				var itemPicked = function(){						
					//	Update the axis setting
					this.settings.predictionMethod = this.caption;
					widget.options.regression = this.settings;
					//	Redraw the widget
					widget.redraw();						
				};

				//	Add all prediction methods
				$.each(mod.predictionMethods,function(){
					
					//	Get the option as a string
					var option = this.label;

					//	Create the menu option
					var item = {
						caption: option,
						checked: (option === selectedMethod),
						size: 'xl',
						type: 'check',
						widget: widget,
						settings: settings,						
						execute: itemPicked
					};

					//	Add to the list
					items.push(item);
				});

				//	Return the list of items				
				return items;
			};

			//	Create header
			var header = {
				caption: headerMenuCaption,
				type: "header",
				disabled: false
			};			

			//	Create menu for picking the regression measure
			var precitionMenuOptions = {
				caption: "Prediction Options",
				items: predictionOptionItems(settings)
			};

			//	Add options to the menu
			args.settings.items.push(header);			
			args.settings.items.push(precitionMenuOptions);

		}
	});

}]);