mod.service("regressionFunc", [
    
        "plugin-RegressionLine.services.predictionMath",
        
    function ($predictionMath) {

        //	Define allowed chart types
        var supportedChartTypes = ["chart/column","chart/line","chart/bar","chart/area","chart/scatter"];



        //	Function to kick off a regression analysis
        this.startRegression = function(widget,args) {

            //	Should we run a quadrant analysis?
            var shouldInit = widgetIsSupported(widget.type);
            //	Add hook for the quadrant analysis
            if (!shouldInit) {
                return;
            }


            //	Get a list of the value items
            var valueItems = getItems(widget,'values');
            var category = getDim(widget);

            if (valueItems) {
                //	Loop through each series
                for (var i=0; i<valueItems.items.length; i++){

                    //	Should regression be calculated for this series?
                    var regressionEnabled = (valueItems.items[i][mod.regressionProperty]) && (valueItems.items[i][mod.regressionProperty] !== mod.none);
                    if (regressionEnabled && !valueItems.items[i].disabled) {

                        //	Get saved settings
                        var settings = widget.options.regression;
                        if (typeof settings === 'undefined'){
                            //	Create defaults, if not defined
                            settings = mod.defaultSettings;
                        }

                        //	Enable regression for this series
                        args.result.series[i].regression = true;

                        //	Define the future periods to predict out
                        var predictionOptions = definePredictionOptions(args.result,args.result.series[i],category,settings);

                        //	Set the options for this series
                        args.result.series[i].regressionSettings = {
                            type: valueItems.items[i][mod.regressionProperty],
                            color: valueItems.items[i].format.color.color,
                            dashStyle: 'dash',
                            name: valueItems.items[i].jaql.title,
                            projections: predictionOptions,
                            style: widget.style
                        };

                        //	Update the chart to show future points
                        addFuturePeriodsAxisLabels(args.result,predictionOptions,i);

                        //	Calculate trendline, and replace the old series
                        var newSeries = calculateTrend(args.result.series[i]);
                        args.result.series[i] = newSeries;
                    }
                }
            }
        };


        //	Function to determine if the regression option should be added
        this.canEnableRegression = function(e,args,headerMenuCaption,menuType){

            //	Assume true by default
            var result = true;
            try {
                //	Has the menu been added already?
                $.each(args.settings.items, function(){
                    if (this.caption === headerMenuCaption) {
                        result = false;
                    }
                });

                //	Only show this in the widget editor
                var widgetEditorOpen = (prism.$ngscope.appstate == "widget");
                if (!widgetEditorOpen) {
                    result = false;
                }

                // find the widget
                var widget = e.currentScope.widget;
                if(widget == undefined){
                    result = false;
                }

                // Widget must be an allowed chart
                if(!widgetIsSupported(widget.type)){
                    result = false;
                }

                //	Check where the user clicked
                var target = args.ui.target;
                if (menuType === 'measure') {
                    //	Make sure the user clicked on a measures menu
                    if (args.settings.name !== "widget-metadataitem") {
                        result = false;
                    }
                    //	Make sure the settings menu is for a value
                    if (args.settings.item.$$panel.name !== "values") {
                        result = false;
                    }
                } else if (menuType === 'settings') {
                    //	Make sure the user clicked on a settings menu
                    if (args.ui.css !== "w-menu-host") {
                        result = false;
                    }
                }


            } catch(e) {
                result = false;
            }

            //	Return result
            return result;
        }


        //	Figure out how many periods to predict out
        function definePredictionOptions(chart,series,dim,settings){

            //	Create options object
            var options = {
                enabled: false,
                predictionMethod: settings.predictionMethod
            };

            //	Is the axis a datetime?
            if (dim.jaql.datatype !== 'datetime' || settings.predictionMethod === mod.none) {

                //	No need to predict into the future, as the axis is not a datetime
                return options;

            } else {

                //	Need to predict into the future
                options.enabled = true;

                //	Get the formatting options
                options.level = dim.jaql.level;
                options.mask = dim.format.mask[dim.jaql.level];

                //	Figure out how far to predict ahead
                var maxDate = series.data[series.data.length-1].selectionData[0];
                var futurePeriods = calculateFuturePeriods(maxDate,options);
                options.projectedPeriods = futurePeriods;

            }

            return options;
        }

        //  Function to determine the # of future periods
        function calculateFuturePeriods(maxDate,options) {

            //	Get parameters
            var predictionMethod = options.predictionMethod;
            var extraPeriods = $.grep(mod.predictionMethods, function(w) {
                return w.label === predictionMethod;
            })[0].numPeriods;
            var level = options.level;

            //  Create an array to send back
            var datesArray = [];

            //	Get the max date info
            var lastMonth = maxDate.getMonth();
            var lastYear = maxDate.getFullYear();
            var lastDay = maxDate.getDate();
            var lastWeek = parseInt(prism.dateFormatter(maxDate,'w'));

            //  Check the level
            if (level === 'days') {

                //  Figure out how many days to predict
                var endDate = new Date(lastYear,lastMonth+1,0);
                var endDay = endDate.getDate();
                var numDays = extraPeriods ? extraPeriods : endDay - lastDay;

                //	Predict out X number of days
                for (var i=1; i<=numDays; i++) {

                    //  Create the new date
                    var newDate = new Date(lastYear,lastMonth,lastDay+i);

                    //  Create the new data object
                    var newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === 'weeks') {

                //  Figure out how many months to predict
                var numWeeks = extraPeriods ? extraPeriods : 52 - lastWeek;

                //  Loop through all remaining dates
                for (var i=1; i<=numWeeks; i++) {

                    //  Create the new date, add 7 days
                    var newDate = new Date(lastYear,lastMonth,lastDay);
                    newDate.setTime( newDate.getTime() + (i*7) * 86400000 );

                    //  Create the new data object
                    var newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === 'months') {

                //  Figure out how many months to predict
                var numMonths = extraPeriods ? extraPeriods : 12 - (lastMonth+1);

                //  Loop through all remaining dates
                for (var i=1; i<=numMonths; i++) {

                    //  Create the new date
                    var newDate = new Date(lastYear,lastMonth+i,1);

                    //  Create the new data object
                    var newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            } else if (level === 'quarters') {

                //  Figure out how many quarters to predict
                var currentQuarter = Math.round((lastMonth+1)/4) + 1;
                var numQuarters = extraPeriods ? extraPeriods : 4;

                //	Loop through all remaining dates
                for (var i=1; i<=numQuarters; i++) {

                    //	Calculate new quarter month index
                    var quarterMonthIndex = (i * 3);

                    //  Create the new date
                    var newDate = new Date(lastYear,lastMonth + quarterMonthIndex,1);

                    //  Create the new data object
                    var newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }

            }  else if (level === 'years') {

                //  Figure out how many years to predict
                var numYears = extraPeriods ? extraPeriods : 5;

                //	Loop through all remaining dates
                for (var i=1; i<=numYears; i++) {

                    //  Create the new date
                    var newDate = new Date(lastYear+i,1,1);

                    //  Create the new data object
                    var newDatapoint = [newDate,null];

                    //  Add to forecast series
                    datesArray.push(newDatapoint);
                }
            }

            return datesArray;
        }

        //	Function to add axis labels
        function addFuturePeriodsAxisLabels(chart,options,index){

            try {
                //	Do we need to run this? true by default
                var shouldRun = true;

                //	Regression not enabled, mark flag as false
                if (options.enabled === false) {
                    shouldRun = false;
                }

                //  Have we already added new dates?
                if (options.periodsAdded) {
                    shouldRun = false;
                }

                //	axis labels already added, mark flag as false
                var chartLength = chart.xAxis.categories.length;
                var dataLength = chart.series[index].data.length + options.projectedPeriods.length;
                if (chartLength === dataLength) {
                    shouldRun = false;
                }

                //	If this should still run, add axis labels
                if (shouldRun){

                    //	Get the formatting mask
                    var mask = options.mask;

                    //	Loop through each point, and create a label
                    $.each(options.projectedPeriods, function(){
                        //	Parse the date
                        //var thisDate = Date.parseDate(this[0]);
                        var thisDate = Date.parseDate ? Date.parseDate(this[0]) : this[0];
                        //	Create a formatted label
                        var label = prism.dateFormatter(thisDate,mask)
                        //	Add to the axis categories
                        chart.xAxis.categories.push(label);
                    })

                    //  specify that the periods have been added
                    options.periodsAdded = true;

                }
            } catch(e) {
                console.log(e);
            }
        }

        //	Function to handle the prediction
        function calculateTrend(oldSeries) {

            //	Reformat the data
            var data = fixData(oldSeries.data);
            var dataOriginal = data.slice(0);

            //	Make sure the regression settings have valid options
            oldSeries.regressionSettings =  oldSeries.regressionSettings || {} ;
            oldSeries.regressionSettings.tooltip = oldSeries.regressionSettings.tooltip || {} ;
            oldSeries.regressionSettings.dashStyle = oldSeries.regressionSettings.dashStyle || 'solid';
            oldSeries.regressionSettings.projections = oldSeries.regressionSettings.projections || {};

            //	Figure out the regression type
            var regressionType = oldSeries.regressionSettings.type || "linear" ;
            var regression;

            //	Create a new data series to show the trend
            var newSeries = {
                data:[],
                color: oldSeries.color ,
                yAxis: oldSeries.yAxis ,
                lineWidth: 2,
                marker: {enabled: false} ,
                isRegressionLine: true,
                type: oldSeries.regressionSettings.linetype || 'spline',
                name: oldSeries.regressionSettings.name || "Equation: %eq",
                color: oldSeries.regressionSettings.color || '',
                dashStyle: oldSeries.regressionSettings.dashStyle || 'solid',
                tooltip:{
                    valueSuffix : oldSeries.regressionSettings.tooltip.valueSuffix || ' '
                }
            };

            //  Figure out which regression type to use
            if (regressionType == "linear") {
                regression = $predictionMath._linear(data) ;
                newSeries.type = "line";
            }else if (regressionType == "exponential") {
                regression = $predictionMath._exponential(data)
            }
            else if (regressionType == "polynomial"){
                var order = s.regressionSettings.order || 2
                regression = $predictionMath._polynomial(data, order) ;
            }else if (regressionType == "logarithmic"){
                regression = $predictionMath._logarithmic(data) ;
            }else if (regressionType == "loess"){
                var loessSmooth = s.regressionSettings.loessSmooth || 25
                regression = $predictionMath._loess(data, loessSmooth/100) ;
            }else {
                console.error("Invalid regression type: " , regressionType) ;
                return;
            }

            //  Predict future data points?
            data = $predictionMath._prediction(regression, oldSeries.regressionSettings.projections,regressionType);

            //	Figure out the r^2 and r values
            regression.rSquared =  $predictionMath.coefficientOfDetermination(dataOriginal, regression.points).toFixed(2);
            regression.rValue = Math.sqrt(regression.rSquared,2).toFixed(2) ;
            newSeries.name = newSeries.name.replace("%r2",regression.rSquared);
            newSeries.name = newSeries.name.replace("%r",regression.rValue);
            newSeries.name = newSeries.name.replace("%eq",regression.string);

            //	Return the data back to object format
            newSeries.data = fixDataReversed(regression.points);

            //	Save the regression results to the data series
            newSeries.regressionOutputs = regression ;

            //	Add the mask and y-axis
            newSeries.mask = oldSeries.mask;
            newSeries.yAxis = oldSeries.yAxis;

            //  Set the user-defined min/max
            /*
             if (oldSeries.regressionSettings) {
             var styler = oldSeries.regressionSettings.style;
             arguments[1].yAxis[0].max = styler.yAxis.max;
             arguments[1].yAxis[0].min = styler.yAxis.min;
             }*/

            //	return the trendline series
            return newSeries;
        }


        //////////////////////////
        //	Utility Functions	//
        //////////////////////////

        //	Function to determine if this chart type is supported
        function widgetIsSupported(type){
            if (supportedChartTypes.indexOf(type) >= 0) {
                return true;
            } else {
                return false;
            }
        }

        //	function to find the value items from a widget
        function getItems(widget,section) {
            var valueItems = $.grep(widget.metadata.panels, function(w){
                return w.title===section;
            })[0];
            return valueItems;
        }

        //	function to find the value items from a widget
        function getDim(widget) {
            //	Get the first panel
            var firstPanel = widget.metadata.panels[0];

            //	Get only the enabled items
            var enabledItems = $.grep(firstPanel.items, function(w) {
                return (typeof w.disabled === 'undefined') || (w.disabled === false);
            });

            //	Return the first enabled item
            return enabledItems[0];
        }

        //	Function to convert the data from its existing structure to the following: [ [dim, value], [dim, value] ]
        function fixData(data){

            //	Create new array to return
            var ret = [data.length];

            //	Loop through the data and structure each object an array of values
            for (var i = 0 ; i < data.length ; i++){
                ret[i] = [2];
                ret[i][0] = data[i].selectionData[0];
                ret[i][1] = data[i].y;
            }

            //	Return new array data
            return ret;
        }

        //	Function to convert the data from an array, back to an objct
        function fixDataReversed(data){

            //	Create a new array to return
            var ret = [data.length];

            //	Loop through the data and convert each array to an object
            for (var i = 0 ; i < data.length ; i++){
                ret[i] = {selectionData : [data[i][0]], y :  data[i][1]};
            }

            //	Return the converted data
            return ret;
        }

    }
]);